% Build a interpolation matrices
% [DOWNMTX, UPMTX] = build_interpolation_fps(D,IND,opts)
%
% inputs:
% D - The input matrix
% IND - The indices of each grid
% opts - An options structure. includes a field 'interType' setting which
% type of interpolation to use:
% - 'uni' - uniform weights
% - 'dist' - 1/d(i,j)
% - 'opt' - solves an optimization problem for each point's weights
% outputs:
% DOWNMTX, UPMTX - the interpolation matrices
%
% for more details, see:
% "G. Rosman, A. M. Bronstein, M. M. Bronstein, R. Kimmel, Manifold
% Analysis by Topologically Constrained Isometric Embedding, World
% Enformatika Conference on Machine Learning and Pattern Recognition (MLPR), 2006
%
% ANY ACADEMIC USE OF THIS CODE MUST CITE THE ABOVE REFERENCES. 
% ANY COMMERCIAL USE PROHIBITED WITHOUT WRITTEN CONSENT OF THE AUTHORS. 
% PLEASE CONTACT THE AUTHORS FOR LICENSING TERMS. 
% PROTECTED BY INTERNATIONAL INTELLECTUAL PROPERTY LAWS AND PATENTS PENDING.
%
function [DOWNMTX, UPMTX] = build_interpolation_fps(D,IND, opts)
if (exist('opts','var')~=1)
    opts=[];
end
if (isfield(opts,'interpType')~=1)
    opts.interpType='opt';
end
interpType=opts.interpType;
R = length(IND);
K = 6; %number of neighbors
if (strcmp(interpType,'opt')==1)
    interpFunc=@(D,d,D2)interpOpt(D,K,D2);
elseif (strcmp(interpType,'dist')==1)
    interpFunc=@(D,d,D2)interpDist(D,d,K);
elseif (strcmp(interpType,'uni')==1)
    interpFunc=@(D,d,D2)interpCent(D,K);
else
    error('wrong interpType');
end

% decimation matrices
DOWNMTX{1} = speye(length(IND{1}), length(IND{1}));
if (length(IND)>1)
    DOWNMTX{1} = DOWNMTX{1}(IND{2},:);
end

for r = 2:R-1,
    DOWNMTX{r} = speye(length(IND{r+1}), length(IND{r}));
end

DOWNMTX{R} = [];


% interpolation matrices
for r = 1:R-1,
   UPMTX{r} = spalloc(length(IND{r}), length(IND{r+1}), length(IND{r+1})); 
   
   % TO DO: remove the loop
   for n = 1:length(IND{r+1}),
       UPMTX{r}(n,n) = 1;
   end
      
   % TO DO: 1. centroid, 2. inv distance 3. MDS
   maxd=[];
   
   for n = length(IND{r+1})+1:length(IND{r}),    
      %iii = find(ADJ(n,:)==1);
       [d,iii] = sort(D(IND{r}(n),IND{r+1}));
       iidx=IND{r+1};
      iii = iii(1:K);
      ii2 = iii(1:min(K,length(iii)));
      
%        w = 1./ones([1 K]);w=w/sum(w);
      Dinc=D([IND{r}(n) iidx(iii)'],[IND{r}(n) iidx(iii)']);
      Dinc2=D([IND{r}(n) iidx(ii2)'],[IND{r}(n) iidx(ii2)']);
      maxd=[maxd,max(Dinc(:))];
      w=interpFunc(Dinc,d,Dinc2);
%       w=interpOpt(Dinc,K);
      
      UPMTX{r}(n,iii) = w;           
   end

end

UPMTX{R} = [];

return

function [w,iter]=getOptimalWeights(Xn,x)
N=size(Xn,2);
opts=optimset('display','off','gradobj','on','Hessian','on','TolX',1E-3,'TolCon',1E-1,'MaxIter',6,'LargeScale','off');
warning off;
[w,fval,exitFlag,output]=fmincon(@fMatch,zeros([N,1]),[],[],ones([1 N]),1,-1*ones([N 1]),1*ones([N 1]),[],opts,Xn,x);
iter=output.iterations;
warning on;
return

function [f,g,h]=fMatch(w,Xn,x)
dif=(Xn*w-x);
f=dif'*dif+1E-6*w'*w;
if nargout==1 
    return;
end
g=2*Xn'*dif+2*1E-6*w;
if nargout==2
    return;
end
h=2*Xn'*Xn+2*(1E-6)*eye(size(Xn,2));
return

function w=interpOpt(Dinc,K,D2)
      Xinc=classicalMds(D2,2,ones(length(D2)))';
      Xneigh=Xinc(:,2:(K+1));
      xchosen=Xinc(:,1);           
      [w,it]=getOptimalWeights(Xneigh,xchosen);
return

function w=interpDist(Dinc,d,K)
w = 1./d(1:K);w=w/sum(w);
return

function w=interpCent(Dinc,K)
w = 1/K*ones([1,K]);
return