% This function performs FPS and yields a hirarchy of grids.
% usage: [IND, EPS,INDB] = fps_grid_weighted(D, W,resolution_vec,initialIdx,boundaryPts)
% Input:  D -  matrix of mutual distances.
%         W - a matrix of pairwise weights
%         resolution_vec - Rx1 vector number of sampled points at each of the R resolution levels.
%         initialIdx - the initial index to start from
%         boundaryPts - a know list of boundary points
%         Note: The number of points is automatically sorted in descending order.  
% 
% Outputs:D_  - Rx1 cell array of mutual distances matrices at different resolutions.
%         IND - Rx1 cell array of vectors of indices of sampled points.
%         EPS - Rx1 array of vectors of sampling radius at each resolution level.
%
% for more details, see:
% "G. Rosman, A. M. Bronstein, M. M. Bronstein, R. Kimmel, Manifold
% Analysis by Topologically Constrained Isometric Embedding, World
% Enformatika Conference on Machine Learning and Pattern Recognition (MLPR), 2006
%
% ANY ACADEMIC USE OF THIS CODE MUST CITE THE ABOVE REFERENCES. 
% ANY COMMERCIAL USE PROHIBITED WITHOUT WRITTEN CONSENT OF THE AUTHORS. 
% PLEASE CONTACT THE AUTHORS FOR LICENSING TERMS. 
% PROTECTED BY INTERNATIONAL INTELLECTUAL PROPERTY LAWS AND PATENTS PENDING.
%
function [IND, EPS,INDB] = fps_grid_weighted(D, W,resolution_vec,initialIdx,boundaryPts)
sorted_resolution_vec = sort (resolution_vec, 'descend');
R = length(sorted_resolution_vec);
if (exist('initialIdx'))
[sampled_ind, epsilon] = fps_weighted(D, W,sorted_resolution_vec(1),boundaryPts,initialIdx);
else
[sampled_ind, epsilon] = fps_weighted(D, W,sorted_resolution_vec(1),boundaryPts);
end
for r = 1:R,
    grid_size=sorted_resolution_vec(r);
    grid_size_original=grid_size;
    IND{r} = sampled_ind(1:grid_size);
    gridStep=10;
    while(~isConnected(D(IND{r},IND{r}),W(IND{r},IND{r})))
        grid_size=min(length(sampled_ind),grid_size+ceil(gridStep));
        gridStep=gridStep*1.5;
        IND{r} = sampled_ind(1:grid_size);
        if (grid_size/grid_size_original>1.5) || (grid_size==length(sampled_ind))
            break;
        end
    end
    D_{r} = D(IND{r},IND{r});
    
    EPS(r) = epsilon(grid_size);
end
% adds some boundary points to each grid, to allow interpolation of
% near-to-boundary points
previous=[];
INDB={};
for r = R:-1:1
    numBoundary=min(length(boundaryPts),ceil(length(IND{r})*0.5));
    IND{r}=[previous; sort(setdiff(unique([boundaryPts(1:numBoundary)';IND{r}(:)]),previous))];
    INDB{r}=find(ismember(IND{r},boundaryPts));
    previous=IND{r};
end
if (length(IND)>1) && ((numel(IND{1})==numel(IND{2})))
    IND={IND{2:end}};
    INDB={INDB{2:end}};
    EPS=EPS(2:end);
end

