% iterate SMACOF iterations using RRE
% D - the distance matrix
% dd - the embedding space dimension
% W - the weights for the LS-stress
% options - contains various options for the optimization
% X - contains the initial matrix of coordinates. Set to empty to allow random initialization
% shouldRRE - If set to 1, uses RRE
% indb - specify boundary indices which should be connected artificially to their neighbors, to prevent disconnected islands
%
% for more details, see:
% "G. Rosman, A. M. Bronstein, M. M. Bronstein, R. Kimmel, Manifold
% Analysis by Topologically Constrained Isometric Embedding, World
% Enformatika Conference on Machine Learning and Pattern Recognition (MLPR), 2006
%
% ANY ACADEMIC USE OF THIS CODE MUST CITE THE ABOVE REFERENCES.
% ANY COMMERCIAL USE PROHIBITED WITHOUT WRITTEN CONSENT OF THE AUTHORS.
% PLEASE CONTACT THE AUTHORS FOR LICENSING TERMS.
% PROTECTED BY INTERNATIONAL INTELLECTUAL PROPERTY LAWS AND PATENTS PENDING.
%

function [X,Xs,stats]=weightedSMACOF(D,dd,W,options,delta,X,shouldRRE,indb)
oldMethod=0;

lambda=0;

[I,J]=meshgrid(1:length(D));
stressIdx=find(J>I);
clear I J
if (isfield(options,'saveStats')~=1) || (options.saveStats==0)
    saveStats=0;
else
    saveStats=1;
end
sumWidx=sum(W(stressIdx));


if (exist('shouldRRE','var')~=1)
    shouldRRE=1;
end
if (isfield(options,'noHistory')==1) && (options.noHistory==1)
    noHistory=1;
else
    noHistory=0;
end
if (isempty(X))
    if (isfield(options,'init') && (strcmp(options.init,'rand')))
        X=rand([length(D),dd]);X=X-0.5;
    else
        X=classicalMds(D,dd,W);
    end
end
if (isfield(options,'ITERS'))
    ITER=options.ITERS;
end
if (isfield(options,'minStress'))
    minStress=options.minStress;
else
    minStress=0;
end
if (isfield(options,'minDStress'))
    minDStress=options.minDStress;
else
    minDStress=0;
end

neighbors=dd+3;
wellConnected=find(sum(W)>0.05*max(W(:)));
indb2=find(sum(W)<=5);
indb2=unique([indb(:)',indb2(:)']);
for i = 1:length(indb2)
    idx=indb2(i);
    [ynn,nearestNeighbors]=sort(D(idx,wellConnected));
    nearestNeighbors=wellConnected(nearestNeighbors(2:neighbors));
    W(idx,nearestNeighbors)=max(W(idx,nearestNeighbors),1);
    W(nearestNeighbors,idx)=max(W(nearestNeighbors,idx),1);
end
N=size(X,1);
K=20;
KN=10;
MAXK=30;
MAXKN=20;
w=1.8; % an over relaxation term
Xss=[];
Rs=[];
Xs={};
Xs{1}=X;
stats.res=[];
stats.cputime=[];
stats.MFLOPS=[];
stats.maxAlgebricError={};
stats.normAlgebricError={};
stats.stress=[];
stats.accelStats={};
stats.meanX1=[];
stats.meanX2=[];
stats.times=[];
stats.extrapolationIdx=[];
stats.postExtrapolationStress=[];
if (isfield(options,'hessianEigenValues') && options.hessianEigenValues==1)
    stats.hessianEV={};
end

if (exist('delta','var')==0)
    delta=0.001;
end
if (exist('ITER','var')==0)
    ITER=10;
end
V=-W;
for i = 1:length(V)
    V(i,i)=0;
    V(i,i)=-sum(V(i,:));
end
% Vp=pinv(V);
if (isfield(options,'count_qr') && options.count_qr==1)
    t1=cputime;
    %mMultCnt=mMultCnt+2/3*length(V)^3;mAddCnt=mAddCnt+2/3*length(V)^3;
end

if (~isfield(options,'count_qr') || (options.count_qr~=1))
    t1=cputime;
end

lastRreIter=0;
retry=0;
for iter=1:ITER

    t2=cputime;
    stats.cputime=[stats.cputime t2-t1];

    % fill the distances matrix
    DX=L2_distance(X',X');
    if (isfield(options,'hessianEigenValues') && options.hessianEigenValues==1)
        [H,e]=calculateHessian(D,X,W,DX);
        stats.hessianEV{end+1}=e;
    end
    addLambda=0;
    if (dd>2)
        addLambda=addLambda+1/length(D)*lambda*sum(X(:,3:end).^2);
    end

    if (isfield(stats,'stress') && (~isempty(stats.stress)))
        oldStress=stats.stress(end);
    end
    if (saveStats)
        stats.stress=[stats.stress calculateStress(X,W,D,stressIdx,sumWidx,addLambda,DX)];
    else
        stats.stress=calculateStress(X,W,D,stressIdx,sumWidx,addLambda,DX);
    end
    if (stats.stress(end)<minStress)
        fprintf('stress=%f, iteration %d, early halt\n',stats.stress(end),iter);
        break;
    end
    if (exist('oldStress','var'))
        dStress=abs(stats.stress(end)-oldStress)/stats.stress(end);
        if (dStress<minDStress)
            break;
        end
    end

    idx=find(DX==0 | eye(N));
    warning off;
    B=-(W.*D./DX);B(idx)=0;
    warning on

    tic
    B=B-diag(sum(B,2));
    BX=B*X;
    if (dd>2)
        BX(:,3:end)=BX(:,3:end)-2/N*lambda*X(:,3:end);
    end

    if (oldMethod)
        Riter=V*X-BX;
        Rs(:,(end+1))=Riter(:);
        [VQ,VR]=qr(V);
        Xnew = VR\(VQ'*BX);
    else
        if (abs(max(W(:))-min(W(:))))/mean((W(:)))<1e-7
            for ddd=1:size(X,2)
                Xnew(:,ddd)=1/size(BX,1)*BX(:,ddd);
            end
        else
            for ddd=1:size(X,2)
                x=X(:,ddd);Xnew(:,ddd)=invertMatrixCG(V+1e-9*eye(length(V)),BX(:,ddd),x);
            end
        end

    end

    Xnew=Xnew-repmat(mean(Xnew),[N, 1]);

    stats.res(end+1)=norm(X(:)-Xnew(:))/N/dd;
    if (norm(Xnew-X,inf)<delta)
        X=Xnew;
        fprintf('Early halt due to small algebric change');
        break;
    end
    stats.maxAlgebricError{end+1}=norm(Xnew-X,inf);
    stats.normAlgebricError{end+1}=norm(Xnew-X,2);
    %     fprintf('.');

    stats.times(end+1)=toc;
    X=w*Xnew+(1-w)*X;
    for i = 1:size(X,2)
        X(:,i)=X(:,i)-mean(X(:,i));
    end
    if (~noHistory)
        Xs{end+1}=X;
    end
    if ((iter-lastRreIter)>=KN)
        Xss=[Xss,X(:)];
    end
    if ((iter-lastRreIter)==K)  && shouldRRE
        stats.extrapolationIdx(end+1)=length(stats.times);
        try
            if (isfield(options,'RREStats')==1) && (options.RREStats==1)
                if (isfield(options,'extrap')==0) ||(isfield(options,'extrap') && (strcmp(options.extrap,'RRE')))
                    [s,accelStat]=RRE(Xss);
                end
                accelStat.linearity=checkSMACOFLinearity(Xss,X,VR,VQ,V,B);
                stats.accelStats{end+1}=accelStat;
            else
                if (isfield(options,'extrap')==0) ||(isfield(options,'extrap') && (strcmp(options.extrap,'RRE')))
                    [s]=RRE(Xss);
                    %                     [s]=MPE(Xss);
                elseif (isfield(options,'extrap') && (strcmp(options.extrap,'WRRE')))
                    deg=2;
                    if (isfield(options,'WRREDEG'))
                        deg=options.WRREDEG;
                    end
                    [s]=RRE3(Xss,deg);
                elseif (isfield(options,'extrap') && (strcmp(options.extrap,'WRRE2')))
                    deg=2;
                    if (isfield(options,'WRREDEG'))
                        deg=options.WRREDEG;
                    end
                    [s]=WRRE(Xss,deg);
                elseif (isfield(options,'extrap') && (strcmp(options.extrap,'RREProj')))
                    [s]=RREproj(Xss,N);
                elseif (isfield(options,'extrap') && (strcmp(options.extrap,'extrap1')))
                    [s]=SMACOFExtrapolate1(Xss,Rs);
                elseif (isfield(options,'extrap') && (strcmp(options.extrap,'extrap2')))
                    [s]=SMACOFExtrapolate2(Xss,Rs,X,D);
                elseif (isfield(options,'extrap') && (strcmp(options.extrap,'extrap3')))
                    [s]=SMACOFExtrapolate3(Xss,Rs);
                end


            end

            KN=2; % no need for preconditioning after the 1st round, usually
            K=6;
            lastRreIter=iter;
            s=reshape(s,size(X));
            s=s-repmat(mean(s),[N, 1]);

            [st,alpha]=lineSearchStress(X,s-X,D,W,stressIdx);
            lsa=0;
            lsc=1;
            if (alpha==0)
                if  (retry)
                    retry=0;
                    K=min(MAXK,K+2);
                    KN=min(MAXKN,KN+1);
                else
                    retry=1;
                    Xss=[];
                    Rs=[];
                    Xnew=X;
                end
            else
                X=st;
                Xss=[];
                Rs=[];
                Xnew=X;
            end


        catch
            lasterr
            fprintf('*');
        end
        if (saveStats)
            stats.postExtrapolationStress(end+1)=calculateStress(X,W,D,stressIdx,sumWidx,addLambda);
        end
    end
    if (isfield(options,'display') && options.display)
        if (size(X,2)==2)
            if (isfield(options,'id'))
                plot(X(:,1),X(:,2),'.',X(options.id,1),X(options.id,2),'r.');axis image;title(sprintf('Iteration %d, stress %8.8f',iter,stats.stress(end)));axis image;drawnow;
            else
                plot(X(:,1),X(:,2),'.');axis image;title(sprintf('Iteration %d, stress %8.8f',iter,stats.stress(end)));axis image;drawnow;
            end
        elseif (size(X,2)==3)
            plot3(X(:,1),X(:,2),X(:,3),'.');axis image;title(sprintf('Iteration %d, stress %8.8f',iter,stats.stress(end)));axis image;drawnow;
        end

    end
    X;
end
stats.yss=Xs;
stats.stressSections={stats.stress};
stats.timeSections={stats.cputime};
stats.flopsSections={stats.MFLOPS};
stats.pts=size(X,1);
end

function s=calculateStress(X,W,D,idx,sumWidx,addLambda,DX)
if (exist('DX','var')==0)
    DX=L2_distance(X',X');
end
s=sum(W(idx).*(D(idx)-DX(idx)).^2)/sumWidx+addLambda;
end