% Usage:
% [X,Xs,stats]=weightedSMACOFMS(D,dimension,W,id,options,delta,X)
% D - the distances matrix
% W - a weights matrix for the least squares stress
% X - contains the initial solution for the coordinates, each row is a
% different point. contains the coordinates at the end
% id - set from which id's of points to start the construction of the
% coarse grid
% options - specifies various options, such as:
% - initNum - the minimum size of the coarse grid
% - RRE - if set to 1, uses RRE
% - minHaltIter - determines the minimum stress change allowed before halting iterations
% delta - specifies the minimum algebraic change allowed before halting iterations
%
% By Guy Rosman, 2008
%
% for more details, see:
% "G. Rosman, A. M. Bronstein, M. M. Bronstein, R. Kimmel, Manifold
% Analysis by Topologically Constrained Isometric Embedding, World
% Enformatika Conference on Machine Learning and Pattern Recognition (MLPR), 2006
%
% ANY ACADEMIC USE OF THIS CODE MUST CITE THE ABOVE REFERENCES.
% ANY COMMERCIAL USE PROHIBITED WITHOUT WRITTEN CONSENT OF THE AUTHORS.
% PLEASE CONTACT THE AUTHORS FOR LICENSING TERMS.
% PROTECTED BY INTERNATIONAL INTELLECTUAL PROPERTY LAWS AND PATENTS PENDING.
%
function [X,Xs,stats]=weightedSMACOFMS(D,dimension,W,id,options,delta,X)
if (exist('options','var')==0)
    options=[];
end
if (isfield(options,'initNum'))
    initNum=options.initNum;
else
    initNum=min(length(D),max(300,min(800,ceil(length(D))/10)));
end
if (exist('id','var')==0)
    id=1;
end
if (isfield(options,'RRE')==1) && (options.RRE==0)
    shouldRRE=0;
else
    shouldRRE=1;
end
if (isfield(options,'minHaltIter'))
    options.minHaltIter=1e-4;
end
% compute multi-scale interpolation grids
CF=2;
resolution_vec=unique(min(length(D),ceil((initNum)*CF.^[0:ceil(log(length(D/initNum))/log(CF)), length(D)])));
[IND, EPS,INDB] = fps_grid_weighted(D,W, resolution_vec,id(1),id);
idx=IND{end};
if (isfield(options,'lengthBias'))
    lengthBias=options.lengthBias;
else
    lengthBias=0;
end
[DOWNMTX, UPMTX] = build_interpolation_fps(D,IND,options);
Widx=W(idx,idx);
idx=IND{end};

if (isfield(options,'graduated'))
    beta=1;
    Widx=Widx/max(Widx(:));Widx=(1-beta)*Widx+(beta)*ones(size(Widx));
elseif (isfield(options,'graduated3'))
    Widx=ones(size(Widx));
elseif (isfield(options,'graduated4'))
    Widx=max(1E-3,Widx);
end
options2=options;
options2.id=find(ismember(idx,id));
if (isfield(options2,'init'))
    options2=rmfield(options2,'init');
end

[Yn,Ys,stats]=weightedSMACOF(D(idx,idx),dimension,((max(max(D(idx,idx)))+D(idx,idx))/2/max(max(D(idx,idx)))).^(lengthBias).*(Widx),options2,delta,[],shouldRRE,INDB{end});% coarsest grid
global ps;
if (isfield(options,'hist')==1)
    ps{end+1}=profile('info');
end
for g=2:length(IND)
    oldIdx=IND{end-g+2};
    idx=IND{end-g+2};
    for ti = min(length(D),(length(oldIdx)+1)):length(IND{end-g+1})
        chosen=IND{end-g+1}(ti);
        upM=UPMTX{end-g+1};
        ychosen=Yn(1:length(oldIdx),:)'*upM(ti,:)';
        Yn=[Yn;ychosen'];
        idx=[idx(:)' chosen];
    end

    Widx=W(idx,idx);
    Widx2=Widx;
    options2=options;
    options2.id=find(ismember(idx,id));
    if (isfield(options,'minDStress')==0)
        options2.minDStress=1e-4;
    end
    if (length(IND)~=g)
        options2.minDStress=max(options2.minDStress,1E-5);
        options2.minStress=0;
    end
    [Yn,Ys2,stats2]=weightedSMACOF(D(idx,idx),dimension,((max(max(D(idx,idx)))+D(idx,idx))/2/max(max(D(idx,idx)))).^(lengthBias*(length(IND)-g)/length(IND)).*Widx2,options2,delta,Yn,shouldRRE,[]);
    %toc
    if (isfield(options,'hist')==1)
        ps{end+1}=profile('info');
    end


    if (isfield(options,'hist')==1)
        stats=combineSmacofStats(stats,stats2);
        Ys={Ys{:},Ys2{:}};
    else
        stats2=stats;
    end
    Y=Yn;
end
stats.idx_mr=idx;
X=zeros(size(Yn));
X(idx,:)=Yn;Xs=Ys;